"""
Trigger collection.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from scheduler.trigger.core import (
    Friday,
    Monday,
    Saturday,
    Sunday,
    Thursday,
    Tuesday,
    Wednesday,
    weekday,
)

__all__ = ["Friday", "Monday", "Saturday", "Sunday", "Thursday", "Tuesday", "Wednesday", "weekday"]
